@@ -18,17 +18,114 @@
     exonEnds="list"       # list of raw vectors
 )
 
-### Lookup between UCSC transcript tables and their associated track.
-.SUPPORTED_UCSC_TABLES <- c(
+### .get_GENCODE_tx_tables("All GENCODE %s", "V34") or
+### .get_GENCODE_tx_tables("GENCODE Genes %s", "V19")
+.get_GENCODE_tx_tables <- function(track, version, only.first.3=FALSE)
+{
+    tx_tables <- c(
+        "wgEncodeGencodeBasic%s",          track, "Basic",
+        "wgEncodeGencodeComp%s",           track, "Comprehensive",
+        "wgEncodeGencodePseudoGene%s",     track, "Pseudogenes"
+    )
+    if (!only.first.3) {
+        tx_tables <- c(tx_tables,
+            "wgEncodeGencode2wayConsPseudo%s", track, "2-way Pseudogenes",
+            "wgEncodeGencodePolya%s",          track, "PolyA"
+        )
+    }
+    sprintf(tx_tables, version)
+}
+
+### NAs in the tablename and track columns are resolved via
+### 'rtracklayer::trackNames(session)'.
+.hg38_TX_TABLES <- c(
+  ## tablename (unique key)           track                subtrack
+  "knownGene",                        NA,                  NA,
+  "ncbiRefSeq",                       "NCBI RefSeq",       "RefSeq All",
+  "ncbiRefSeqCurated",                "NCBI RefSeq",       "RefSeq Curated",
+  "ncbiRefSeqPredicted",              "NCBI RefSeq",       "RefSeq Predicted",
+  "ncbiRefSeqOther",                  "NCBI RefSeq",       "RefSeq Other",
+  #"ncbiRefSeqPsl",                    "NCBI RefSeq",       "RefSeq Alignments",
+  #"ncbiRefSeqGenomicDiff",            "NCBI RefSeq",       "RefSeq Diffs",
+  "refGene",                          "NCBI RefSeq",       "UCSC RefSeq",
+  "ncbiRefSeqSelect",                 "NCBI RefSeq",       "RefSeq Select+MANE",
+  "ncbiRefSeqHgmd",                   "NCBI RefSeq",       "RefSeq HGMD",
+  "xenoRefGene",                      "Other RefSeq",      NA,
+  .get_GENCODE_tx_tables("All GENCODE %s", "V34"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V33"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V31"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V29"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V28"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V27"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V26"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V25"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V24"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V23"),
+  .get_GENCODE_tx_tables("All GENCODE %s", "V22"),
+  .get_GENCODE_tx_tables("GENCODE %s (Ensembl 76)", "V20"),
+  "augustusGene",                     "AUGUSTUS",          NA,
+  "ccdsGene",                         "CCDS",              NA,
+  "geneid",                           "Geneid Genes",      NA,
+  "genscan",                          "Genscan Genes",     NA,
+  NA,                                 "Old UCSC Genes",    NA,
+  "sgpGene",                          "SGP Genes",         NA,
+  "sibGene",                          "SIB Genes",         NA
+)
+
+### NAs in the tablename and track columns are resolved via
+### 'rtracklayer::trackNames(session)'.
+.hg19_TX_TABLES <- c(
+  ## tablename (unique key)           track                subtrack
+  "knownGene",                        NA,                  NA,
+  "ncbiRefSeq",                       "NCBI RefSeq",       "RefSeq All",
+  "ncbiRefSeqCurated",                "NCBI RefSeq",       "RefSeq Curated",
+  "ncbiRefSeqPredicted",              "NCBI RefSeq",       "RefSeq Predicted",
+  "ncbiRefSeqOther",                  "NCBI RefSeq",       "RefSeq Other",
+  #"ncbiRefSeqPsl",                    "NCBI RefSeq",       "RefSeq Alignments",
+  #"ncbiRefSeqGenomicDiff",            "NCBI RefSeq",       "RefSeq Diffs",
+  "refGene",                          "NCBI RefSeq",       "UCSC RefSeq",
+  "ncbiRefSeqSelect",                 "NCBI RefSeq",       "RefSeq Select",
+  "ncbiRefSeqHgmd",                   "NCBI RefSeq",       "RefSeq HGMD",
+  "xenoRefGene",                      "Other RefSeq",      NA,
+  "acembly",                          "AceView Genes",     NA,
+  "augustusGene",                     "AUGUSTUS",          NA,
+  "ccdsGene",                         "CCDS",              NA,
+  "ensGene",                          "Ensembl Genes",     NA,
+  "exoniphy",                         "Exoniphy",          NA,
+  .get_GENCODE_tx_tables("GENCODE %s", "V34lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE %s", "V33lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE %s", "V31lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE %s", "V28lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE Gene %s", "V27lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE Gene %s", "V24lift37", only.first.3=TRUE),
+  .get_GENCODE_tx_tables("GENCODE Genes %s", "V19"),
+  .get_GENCODE_tx_tables("GENCODE Genes %s", "V17"),
+  .get_GENCODE_tx_tables("GENCODE Genes %s", "V14"),
+  .get_GENCODE_tx_tables("GENCODE Genes %s", "V7"),
+  "geneid",                           "Geneid Genes",      NA,
+  "genscan",                          "Genscan Genes",     NA,
+  "nscanGene",                        "N-SCAN",            NA,
+  NA,                                 "Old UCSC Genes",    NA,
+  "sgpGene",                          "SGP Genes",         NA,
+  "sibGene",                          "SIB Genes",         NA,
+  "vegaGene",                         "Vega Genes",        "Vega Protein Genes",
+  "vegaPseudoGene",                   "Vega Genes",        "Vega Pseudogenes",
+  "pseudoYale60",                     "Yale Pseudo60",     NA
+)
+
+### NAs in the tablename and track columns are resolved via
+### 'rtracklayer::trackNames(session)'.
+.ALL_SUPPORTED_TX_TABLES <- c(
   ## tablename (unique key)           track                subtrack
 
-  ## Tables/tracks shared by hg18/hg19.
+  ## Tables/tracks for hg18.
   ## All the tables/tracks listed in this section belong to the "Genes and
-  ## Gene Prediction" group of tracks for hg18, hg19, hg38, mm10, and sacCer2.
+  ## Gene Prediction" group of tracks for hg18, mm10, and sacCer2.
   ## On Aug 13 2010, makeTxDbFromUCSC() was successfully tested by hand on
   ## all of them for hg18 (i.e. with 'genome="hg18"').
   ## Note: the "acembly" table contains more than 250000 transcripts!
-  "knownGene",                        "UCSC Genes",        NA,
+  "knownGene",                        NA,                  NA,
+  "knownGeneOld11",                   "Old UCSC Genes",    NA,
   "knownGeneOld8",                    "Old UCSC Genes",    NA,
   "knownGeneOld7",                    "Old UCSC Genes",    NA,
   "knownGeneOld6",                    "Old UCSC Genes",    NA,
@@ -36,7 +133,7 @@
   "knownGeneOld3",                    "Old UCSC Genes",    NA,
   "knownGenePrevious",                "Old Known Genes",   NA,
   "ccdsGene",                         "CCDS",              NA,
-  "refGene",                          "RefSeq Genes",      NA,
+  "refGene",                          "UCSC RefSeq",       NA,
   "xenoRefGene",                      "Other RefSeq",      NA,
   "vegaGene",                         "Vega Genes",        "Vega Protein Genes",
   "vegaPseudoGene",                   "Vega Genes",        "Vega Pseudogenes",
@@ -49,7 +146,7 @@
   "geneid",                           "Geneid Genes",      NA,
   "genscan",                          "Genscan Genes",     NA,
   "exoniphy",                         "Exoniphy",          NA,
-  "augustusGene",                     "Augustus",          NA,
+  "augustusGene",                     "AUGUSTUS",          NA,
   "augustusHints",                    "Augustus",          "Augustus Hints",
   "augustusXRA",                      "Augustus",          "Augustus De Novo",
   "augustusAbinitio",                 "Augustus",          "Augustus Ab Initio",
@@ -61,42 +158,18 @@
   "wgEncodeGencodeAutoV3",            "Gencode Genes",     "Gencode Auto",
   "wgEncodeGencodePolyaV3",           "Gencode Genes",     "Gencode PolyA",
 
-  ## Tables/tracks specific to hg19.
-  "wgEncodeGencodeBasicV19",          "GENCODE Genes V19", NA,
-  "wgEncodeGencodeCompV19",           "GENCODE Genes V19", NA,
-  "wgEncodeGencodePseudoGeneV19",     "GENCODE Genes V19", NA,
-  "wgEncodeGencode2wayConsPseudoV19", "GENCODE Genes V19", NA,
-  "wgEncodeGencodePolyaV19",          "GENCODE Genes V19", NA,
-
-  "wgEncodeGencodeBasicV17",          "GENCODE Genes V17", NA,
-  "wgEncodeGencodeCompV17",           "GENCODE Genes V17", NA,
-  "wgEncodeGencodePseudoGeneV17",     "GENCODE Genes V17", NA,
-  "wgEncodeGencode2wayConsPseudoV17", "GENCODE Genes V17", NA,
-  "wgEncodeGencodePolyaV17",          "GENCODE Genes V17", NA,
-
-  "wgEncodeGencodeBasicV14",          "GENCODE Genes V14", NA,
-  "wgEncodeGencodeCompV14",           "GENCODE Genes V14", NA,
-  "wgEncodeGencodePseudoGeneV14",     "GENCODE Genes V14", NA,
-  "wgEncodeGencode2wayConsPseudoV14", "GENCODE Genes V14", NA,
-  "wgEncodeGencodePolyaV14",          "GENCODE Genes V14", NA,
-
-  "wgEncodeGencodeBasicV7",           "GENCODE Genes V7",  NA,
-  "wgEncodeGencodeCompV7",            "GENCODE Genes V7",  NA,
-  "wgEncodeGencodePseudoGeneV7" ,     "GENCODE Genes V7",  NA,
-  "wgEncodeGencode2wayConsPseudoV7",  "GENCODE Genes V7",  NA,
-  "wgEncodeGencodePolyaV7",           "GENCODE Genes V7",  NA,
-
-  ## Tables/tracks specific to hg38/hg19/mm10/rn6/danRer10/danRer11/ce11/dm6/sacCer3.
+  ## Tables/tracks specific to mm10/rn6/danRer10/danRer11/ce11/dm6/sacCer3.
   "ncbiRefSeq",                       "NCBI RefSeq",       "RefSeq All",
   "ncbiRefSeqCurated",                "NCBI RefSeq",       "RefSeq Curated",
-  "ncbiRefSeqOther",                  "NCBI RefSeq",       "RefSeq Other",
-  "ncbiRefSeqPsl",                    "NCBI RefSeq",       "RefSeq Alignments",
-
-  ## Tables/tracks specific to hg38/mm10/rn6/danRer10/danRer11.
   "ncbiRefSeqPredicted",              "NCBI RefSeq",       "RefSeq Predicted",
+  "ncbiRefSeqOther",                  "NCBI RefSeq",       "RefSeq Other",
+  #"ncbiRefSeqPsl",                    "NCBI RefSeq",       "RefSeq Alignments",
+  #"ncbiRefSeqGenomicDiff",            "NCBI RefSeq",       "RefSeq Diffs",
+  "ncbiRefSeqSelect",                 "NCBI RefSeq",       "RefSeq Select+MANE",
+  "ncbiRefSeqHgmd",                   "NCBI RefSeq",       "RefSeq HGMD",
 
-  ## Tables/tracks specific to hg38.
-  "ncbiRefSeqGenomicDiff",            "NCBI RefSeq",       "RefSeq Diffs",
+  ## Tables/tracks specific to mm10/rn6/danRer10/danRer11.
+  "ncbiRefSeqPredicted",              "NCBI RefSeq",       "RefSeq Predicted",
 
   ## Tables/tracks specific to D. melanogaster.
   "flyBaseGene",                      "FlyBase Genes",     NA,
@@ -107,12 +180,25 @@
   "sgdGene",                          "SGD Genes",         NA
 )
 
+.get_supported_tx_tables <- function(genome)
+{
+    tx_tables <- switch(genome,
+        hg38=.hg38_TX_TABLES,
+        hg19=.hg19_TX_TABLES,
+        .ALL_SUPPORTED_TX_TABLES
+    )
+    m <- matrix(tx_tables, ncol=3L, byrow=TRUE)
+    colnames(m) <- c("tablename", "track", "subtrack")
+    as.data.frame(m)
+}
+
 ### Return a data.frame with 3 columns (tablename, track, and subtrack) and
 ### 1 row per UCSC table known to work with makeTxDbFromUCSC().
 ### A note about the current implementation:
-### Current implementation uses hard-coded .SUPPORTED_UCSC_TABLES matrix
-### above which is not satisfying in the long run (the matrix needs to be
-### manually updated from times to times, a long and boring and error-prone
+### Current implementation uses hard-coded matrices .hg38_TX_TABLES,
+### .hg19_TX_TABLES, and .ALL_SUPPORTED_TX_TABLES defined above which
+### is not satisfying in the long run (the matrices need to be manually
+### updated from times to times, a long and boring and error-prone
 ### process, and is probably out-of-sync at the moment). Ideally we'd like
 ### to be able to generate the 3-column data.frame programmatically in
 ### reasonable time. For this we need to be able to retrieve all the "central
@@ -126,49 +212,19 @@ supportedUCSCtables <- function(genome="hg19",
     if (is(genome, "UCSCSession")) {
         if (!missing(url))
             warning("'url' is ignored when 'genome' is a UCSCSession object")
-    } else {
-        if (!isSingleStringOrNA(genome))
-            stop("'genome' must be a single string or NA")
-        if (!isSingleString(url))
-            stop("'url' must be a single string")
-    }
-    mat <- matrix(.SUPPORTED_UCSC_TABLES, ncol=3, byrow=TRUE)
-    colnames(mat) <- c("tablename", "track", "subtrack")
-    ans_tablename <- mat[ , "tablename"]
-    ans_track <- mat[ , "track"]
-    ans_subtrack <- mat[ , "subtrack"]
-    ans <- data.frame(tablename=ans_tablename,
-                      track=ans_track,
-                      subtrack=ans_subtrack,
-                      stringsAsFactors=FALSE)
-    if (isSingleStringOrNA(genome) && is.na(genome)) {
-        ans$track <- factor(ans_track, levels=unique(ans_track))
-        return(ans)
-    }
-    if (is(genome, "UCSCSession")) {
         session <- genome
         genome <- genome(session)
     } else {
+        if (!isSingleString(genome))
+            stop("'genome' must be a single string")
+        if (!isSingleString(url))
+            stop("'url' must be a single string")
         session <- browserSession(url=url)
         genome(session) <- genome
     }
-    if (genome %in% c("hg17", "hg16", "mm8", "mm7", "rn3")) {
-        ans_track[ans$tablename == "knownGene"] <- "Known Genes"
-        ans$track <- ans_track
-    } else if (genome %in% c("hg38", "hg19", "mm10", "rn6",
-                             "bosTau9", "danRer10", "danRer11",
-                             "ce11", "dm6", "galGal6", "panTro6",
-                             "rheMac10", "sacCer3"))
-    {
-        if (genome %in% "hg38")
-            ans_track[ans$tablename == "knownGene"] <- "GENCODE v29"
-        if (genome %in% "mm10")
-            ans_track[ans$tablename == "knownGene"] <- "GENCODE VM20"
-        ans_track[ans$tablename == "refGene"] <- "NCBI RefSeq"
-        ans_subtrack[ans$tablename == "refGene"] <- "UCSC RefSeq"
-        ans$track <- ans_track
-        ans$subtrack <- ans_subtrack
-    }
+
+    ans <- .get_supported_tx_tables(genome)
+
     ## trackNames() returns a mapping from track names to "central table" names
     ## in the form of a named character vector where the names are the track
     ## names and the values the "central table" names (more than 1 table can
@@ -182,9 +238,36 @@ supportedUCSCtables <- function(genome="hg19",
     ## track=track)) in a loop on all the tracks returned by trackNames()
     ## would work but is very slow :-/
     genome_tracknames <- trackNames(session)
-    ans <- ans[ans$track %in% names(genome_tracknames), , drop=FALSE]
+
+    ## Resolve NAs in the tablename column.
+    na_idx <- which(is.na(ans[ , "tablename"]))
+    m <- match(ans[na_idx, "track"], names(genome_tracknames))
+    ans[na_idx, "tablename"] <- genome_tracknames[m]
+
+    ## Resolve NAs in the track column.
+    na_idx <- which(is.na(ans[ , "track"]))
+    m <- match(ans[na_idx, "tablename"], genome_tracknames)
+    ans[na_idx, "track"] <- names(genome_tracknames)[m]
+
+    if (!(genome %in% c("hg38", "hg19"))) {
+        ## Keep only existing tracks.
+        ans <- ans[ans$track %in% names(genome_tracknames), , drop=FALSE]
+        rownames(ans) <- NULL
+
+        ## Associate subtrack "UCSC RefSeq" to table "refGene" for a few
+        ## genome builds.
+        if (genome %in% c("mm10", "rn6",
+                          "bosTau9", "danRer10", "danRer11",
+                          "ce11", "dm6", "galGal6", "panTro6",
+                          "rheMac10", "sacCer3"))
+        {
+            ans_subtrack <- ans[ , "subtrack"]
+            ans_subtrack[ans[ , "tablename"] == "refGene"] <- "UCSC RefSeq"
+            ans[ , "subtrack"] <- ans_subtrack
+        }
+    }
+
     ans$track <- factor(ans$track, levels=unique(ans$track))
-    rownames(ans) <- NULL
     ans
 }
 
@@ -448,7 +531,7 @@ browseUCSCtrack <- function(genome="hg19",
     )
 )
 
-.howToGetTxName2GeneIdMapping <- function(tablename)
+.get_txname2geneid_mapdef <- function(tablename)
     .UCSC_TXNAME2GENEID_MAPDEFS[[tablename]]
 
 .fetch_UCSC_txtable <- function(genome, tablename, transcript_ids=NULL)
@@ -460,7 +543,7 @@ browseUCSCtrack <- function(genome="hg19",
                          paste(paste0("'", transcript_ids, "'"), collapse=","))
     }
     columns <- names(.UCSC_TXCOL2CLASS)
-    mapdef <- .howToGetTxName2GeneIdMapping(tablename)
+    mapdef <- .get_txname2geneid_mapdef(tablename)
     if (is.character(mapdef))
         columns <- c(columns, mapdef[["colname"]])
     message("Download the ", tablename, " table ... ", appendLF=FALSE)
@@ -483,19 +566,19 @@ browseUCSCtrack <- function(genome="hg19",
     ucsc_txtable
 }
 
-.fetch_UCSC_table <- function(session, tablename, columns=NULL)
+.fetch_UCSC_table <- function(genome, tablename, columns=NULL)
 {
     message("Download the ", tablename, " table ... ", appendLF=FALSE)
     on.exit(message("OK"))
     if (tablename == "hgFixed.refLink")
         return(UCSC_dbselect("hgFixed", "refLink", columns=columns))
-    UCSC_dbselect(genome(session), tablename, columns=columns)
+    UCSC_dbselect(genome, tablename, columns=columns)
 }
 
 ### The 2 functions below must return a named list with 2 elements:
 ###   $genes: data.frame with tx_name and gene_id cols;
 ###   $gene_id_type: single string.
-.fetchTxName2GeneIdMappingFromUCSC <- function(session, Ltablename, mapdef)
+.fetch_txname2geneid_from_UCSC <- function(genome, Ltablename, mapdef)
 {
     nlink <- length(mapdef$L2Rchain)
     for (i in seq_len(nlink)) {
@@ -508,7 +591,7 @@ browseUCSCtrack <- function(genome="hg19",
         ## "wgEncodeGencodeAttrsV17" table does NOT belong to the "GENCODE
         ## Genes V17" track) so we don't need the track information to fetch
         ## these tables.
-        ucsc_table <- .fetch_UCSC_table(session, tablename,
+        ucsc_table <- .fetch_UCSC_table(genome, tablename,
                                         columns=c(Lcolname, Rcolname))
         if (!all(has_col(ucsc_table, c(Lcolname, Rcolname))))
             stop("expected cols \"", Lcolname, "\" or/and \"",
@@ -534,7 +617,7 @@ browseUCSCtrack <- function(genome="hg19",
     list(genes=genes, gene_id_type=gene_id_type)
 }
 
-.extractTxName2GeneIdMappingFromUCSCTxTable <- function(ucsc_txtable, mapdef)
+.extract_txname2geneid_from_UCSC_txtable <- function(ucsc_txtable, mapdef)
 {
     genes <- data.frame(tx_name=ucsc_txtable[["name"]],
                         gene_id=ucsc_txtable[[mapdef[["colname"]]]],
@@ -548,7 +631,7 @@ browseUCSCtrack <- function(genome="hg19",
 ### Extract the 'transcripts' data frame from UCSC table.
 ###
 
-.extractTranscriptsFromUCSCTxTable <- function(ucsc_txtable)
+.extract_transcripts_from_UCSC_txtable <- function(ucsc_txtable)
 {
     message("Extract the 'transcripts' data frame ... ", appendLF=FALSE)
     on.exit(message("OK"))
@@ -604,8 +687,8 @@ browseUCSCtrack <- function(genome="hg19",
 ###       have the cdsStartStat/cdsStartEnd columns). For hg19 ccdsGene, this
 ###       is not true anymore :-/
 ### TODO: Investigate (1) and (2).
-.extractUCSCCdsStartEnd <- function(cdsStart0, cdsEnd1,
-                                    exon_start0, exon_end1, tx_name)
+.extract_UCSC_cds_start_end <- function(cdsStart0, cdsEnd1,
+                                        exon_start0, exon_end1, tx_name)
 {
     cds_start0 <- cds_end1 <- integer(length(exon_start0))
     cds_start0[] <- NA_integer_
@@ -643,13 +726,13 @@ browseUCSCtrack <- function(genome="hg19",
 ### of integer vectors with eventually NAs. The 2 elements have the same
 ### "shape" as ucsc_txtable$exonStarts and ucsc_txtable$exonEnds and the NAs
 ### in them occur at the same places in the 2 elements.
-.extractCdsLocsFromUCSCTxTable <- function(ucsc_txtable)
+.extract_cds_locs_from_UCSC_txtable <- function(ucsc_txtable)
 {
-    startend <- Map(.extractUCSCCdsStartEnd,
-                    ucsc_txtable$cdsStart, ucsc_txtable$cdsEnd,
-                    ucsc_txtable$exonStarts, ucsc_txtable$exonEnds,
-                    ucsc_txtable$name)
-    bad <- sapply(startend, "[[", 3L)
+    start_end <- Map(.extract_UCSC_cds_start_end,
+                     ucsc_txtable$cdsStart, ucsc_txtable$cdsEnd,
+                     ucsc_txtable$exonStarts, ucsc_txtable$exonEnds,
+                     ucsc_txtable$name)
+    bad <- sapply(start_end, "[[", 3L)
     if (any(bad)) {
         bad_cds <- ucsc_txtable$name[bad]
         msg <- sprintf("UCSC data anomaly in %d transcript(s):
@@ -658,11 +741,12 @@ browseUCSCtrack <- function(genome="hg19",
             paste(sQuote(bad_cds), collapse=" "))
         warning(paste(strwrap(msg, exdent=2L), collapse="\n"))
     }
-    list(start=sapply(startend, "[[", 1L),
-         end=sapply(startend, "[[", 2L))
+    list(start=sapply(start_end, "[[", 1L),
+         end=sapply(start_end, "[[", 2L))
 }
 
-.extractSplicingsFromUCSCTxTable <- function(ucsc_txtable, transcripts_tx_id)
+.extract_splicings_from_UCSC_txtable <- function(ucsc_txtable,
+                                                 transcripts_tx_id)
 {
     message("Extract the 'splicings' data frame ... ", appendLF=FALSE)
     on.exit(message("OK"))
@@ -675,8 +759,8 @@ browseUCSCtrack <- function(genome="hg19",
         if (min(exon_count) <= 0L)
             stop("UCSC data anomaly: 'ucsc_txtable$exonCount' contains ",
                  "non-positive values")
-        exon_rank <- makeExonRankCol(exon_count, ucsc_txtable$strand)
-        cds_locs <- .extractCdsLocsFromUCSCTxTable(ucsc_txtable)
+        exon_rank <- make_exon_rank_col(exon_count, ucsc_txtable$strand)
+        cds_locs <- .extract_cds_locs_from_UCSC_txtable(ucsc_txtable)
         exon_start <- unlist(ucsc_txtable$exonStarts) + 1L
         exon_end <- unlist(ucsc_txtable$exonEnds)
         cds_start <- unlist(cds_locs$start) + 1L
@@ -698,7 +782,7 @@ browseUCSCtrack <- function(genome="hg19",
 ### Preprocess the 'genes' data frame.
 ###
 
-.makeUCSCGenes <- function(genes, ucsc_txtable)
+.make_UCSC_genes <- function(genes, ucsc_txtable)
 {
     #genes <- S4Vectors:::extract_data_frame_rows(genes,
     #                             genes$tx_name %in% ucsc_txtable$name)
@@ -710,49 +794,66 @@ browseUCSCtrack <- function(genome="hg19",
 ### Download and preprocess the 'chrominfo' data frame.
 ###
 
-.makeUCSCChrominfo <- function(genome, circ_seqs=DEFAULT_CIRC_SEQS,
-        goldenPath_url="http://hgdownload.cse.ucsc.edu/goldenPath")
+.warn_about_circularity_guess <- function(genome, circ_seqs)
+{
+    if (length(circ_seqs) == 0L) {
+        guess <- c("no sequence in ", genome , " is circular")
+    } else if (length(circ_seqs) == 1L) {
+        guess <- c(circ_seqs, " is circular")
+    } else {
+        in1string <- paste(circ_seqs, collapse=", ")
+        guess <- c(length(circ_seqs), " sequences (",
+                   in1string, ") are circular")
+    }
+    warning(wmsg("UCSC genome ", genome, " is not registered in ",
+                 "GenomeInfoDb (see ?registered_UCSC_genomes), so ",
+                 "we made the following guess about this assembly: ",
+                 guess, ". You can see this by calling isCircular() ",
+                 "on the TxDb object returned by makeTxDbFromUCSC(). "),
+            "\n  ",
+            wmsg("If you think this is incorrect, please let us know. ",
+                 "In the meantime you can supply your own list of ",
+                 "circular sequences (as a character vector) via ",
+                 "the 'circ_seqs' argument."))
+}
+
+.make_UCSC_chrominfo <- function(genome, circ_seqs=NULL,
+        goldenPath.url=getOption("UCSC.goldenPath.url"))
 {
     message("Download and preprocess the 'chrominfo' data frame ... ",
             appendLF=FALSE)
     on.exit(message("OK"))
-    ucsc_chrominfotable <- GenomeInfoDb:::fetch_ChromInfo_from_UCSC(genome,
-                                                goldenPath_url)
-    COL2CLASS <- c(
-        chrom="character",
-        size="integer"
-    )
-    ucsc_chrominfotable <- setDataFrameColClass(ucsc_chrominfotable, COL2CLASS,
-                                                drop.extra.cols=TRUE)
+    chrom_info <- getChromInfoFromUCSC(genome, goldenPath.url=goldenPath.url)
+    if (!is.null(circ_seqs)) {
+        chrom_info[ , "circular"] <-
+            GenomeInfoDb:::make_circ_flags_from_circ_seqs(
+                                           chrom_info[ , "chrom"],
+                                           circ_seqs=circ_seqs)
+    } else {
+        UCSC_genomes <- registered_UCSC_genomes()[ , "genome"]
+        if (!(genome %in% UCSC_genomes)) {
+            circ_seqs <- circ_seqs[circ_seqs[ , "circular"], "chrom"]
+            .warn_about_circularity_guess(genome, circ_seqs)
+        }
+    }
     chrominfo <- data.frame(
-        chrom=ucsc_chrominfotable$chrom,
-        length=ucsc_chrominfotable$size,
-        is_circular=make_circ_flags_from_circ_seqs(ucsc_chrominfotable$chrom,
-                                                   circ_seqs),
+        chrom=chrom_info[ , "chrom"],
+        length=chrom_info[ , "size"],
+        is_circular=chrom_info[ , "circular"],
         stringsAsFactors=FALSE
     )
     oo <- order(rankSeqlevels(chrominfo[ , "chrom"]))
     S4Vectors:::extract_data_frame_rows(chrominfo, oo)
 }
 
-## User-friendly wrapper to .makeUCSCChrominfo().
-getChromInfoFromUCSC <- function(genome,
-          goldenPath_url="http://hgdownload.cse.ucsc.edu/goldenPath")
-{
-  chrominfo <-.makeUCSCChrominfo(genome, circ_seqs=character(),
-                                 goldenPath_url=goldenPath_url)
-  chrominfo[ , 1:2]
-}
-
-
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Prepare the 'metadata' data frame.
 ###
 
-.prepareUCSCMetadata <- function(genome, tablename, track, gene_id_type,
-                                 full_dataset,
-                                 taxonomyId=NA, miRBaseBuild=NA)
+.prepare_UCSC_metadata <- function(genome, tablename, track, gene_id_type,
+                                   full_dataset,
+                                   taxonomyId=NA, miRBaseBuild=NA)
 {
     message("Prepare the 'metadata' data frame ... ", appendLF=FALSE)
     on.exit(message("OK"))
@@ -785,11 +886,11 @@ getChromInfoFromUCSC <- function(genome,
 ### makeTxDbFromUCSC()
 ###
 
-.makeTxDbFromUCSCTxTable <- function(ucsc_txtable, genes,
+.make_TxDb_from_UCSC_txtable <- function(ucsc_txtable, genes,
         genome, tablename, track, gene_id_type,
         full_dataset,
-        circ_seqs=DEFAULT_CIRC_SEQS,
-        goldenPath_url="http://hgdownload.cse.ucsc.edu/goldenPath",
+        circ_seqs=NULL,
+        goldenPath.url=getOption("UCSC.goldenPath.url"),
         taxonomyId=NA,
         miRBaseBuild=NA)
 {
@@ -803,14 +904,14 @@ getChromInfoFromUCSC <- function(genome,
                                                             keep_idx)
     }
 
-    transcripts <- .extractTranscriptsFromUCSCTxTable(ucsc_txtable)
-    splicings <- .extractSplicingsFromUCSCTxTable(ucsc_txtable,
-                                                  transcripts$tx_id)
-    genes <- .makeUCSCGenes(genes, ucsc_txtable)
-    chrominfo <- .makeUCSCChrominfo(genome, circ_seqs, goldenPath_url)
-    metadata <- .prepareUCSCMetadata(genome, tablename, track, gene_id_type,
-                                     full_dataset,
-                                     taxonomyId,  miRBaseBuild)
+    transcripts <- .extract_transcripts_from_UCSC_txtable(ucsc_txtable)
+    splicings <- .extract_splicings_from_UCSC_txtable(ucsc_txtable,
+                                                      transcripts$tx_id)
+    genes <- .make_UCSC_genes(genes, ucsc_txtable)
+    chrominfo <- .make_UCSC_chrominfo(genome, circ_seqs, goldenPath.url)
+    metadata <- .prepare_UCSC_metadata(genome, tablename, track, gene_id_type,
+                                       full_dataset,
+                                       taxonomyId,  miRBaseBuild)
     ## Jan 2019 -- The refGene tables in the hg19 and hg38 UCSC databases were
     ## last updated in Nov 2018 and now contain transcripts located on
     ## sequences that don't belong to the underlying genomes (GRCh37 and GRCh38
@@ -854,9 +955,9 @@ getChromInfoFromUCSC <- function(genome,
 makeTxDbFromUCSC <- function(genome="hg19",
         tablename="knownGene",
         transcript_ids=NULL,
-        circ_seqs=DEFAULT_CIRC_SEQS,
+        circ_seqs=NULL,
         url="http://genome.ucsc.edu/cgi-bin/",
-        goldenPath_url="http://hgdownload.cse.ucsc.edu/goldenPath",
+        goldenPath.url=getOption("UCSC.goldenPath.url"),
         taxonomyId=NA,
         miRBaseBuild=NA)
 {
@@ -871,8 +972,8 @@ makeTxDbFromUCSC <- function(genome="hg19",
     }
     if (!isSingleString(url))
         stop("'url' must be a single string")
-    if (!isSingleString(goldenPath_url))
-        stop("'goldenPath_url' must be a single string")
+    if (!isSingleString(goldenPath.url))
+        stop("'goldenPath.url' must be a single string")
 
     ## Create an UCSC Genome Browser session.
     session <- browserSession(url=url)
@@ -884,25 +985,26 @@ makeTxDbFromUCSC <- function(genome="hg19",
                                         transcript_ids=transcript_ids)
 
     ## Get the tx_name-to-gene_id mapping.
-    mapdef <- .howToGetTxName2GeneIdMapping(tablename)
+    mapdef <- .get_txname2geneid_mapdef(tablename)
     if (is.null(mapdef)) {
         txname2geneid <- list(genes=NULL, gene_id_type="no gene ids")
     } else if (is.list(mapdef)) {
-        txname2geneid <- .fetchTxName2GeneIdMappingFromUCSC(session,
+        txname2geneid <- .fetch_txname2geneid_from_UCSC(
+                                 genome(session),
                                  tablename, mapdef)
     } else if (is.character(mapdef)) {
-        txname2geneid <- .extractTxName2GeneIdMappingFromUCSCTxTable(
+        txname2geneid <- .extract_txname2geneid_from_UCSC_txtable(
                                  ucsc_txtable, mapdef)
     } else {
         stop("GenomicFeatures internal error: invalid 'mapdef'")
     }
-    .makeTxDbFromUCSCTxTable(ucsc_txtable, txname2geneid$genes,
-                             genome, tablename, track,
-                             txname2geneid$gene_id_type,
-                             full_dataset=is.null(transcript_ids),
-                             circ_seqs=circ_seqs,
-                             goldenPath_url=goldenPath_url,
-                             taxonomyId=taxonomyId,
-                             miRBaseBuild=miRBaseBuild)
+    .make_TxDb_from_UCSC_txtable(ucsc_txtable, txname2geneid$genes,
+                                 genome, tablename, track,
+                                 txname2geneid$gene_id_type,
+                                 full_dataset=is.null(transcript_ids),
+                                 circ_seqs=circ_seqs,
+                                 goldenPath.url=goldenPath.url,
+                                 taxonomyId=taxonomyId,
+                                 miRBaseBuild=miRBaseBuild)
 }
 
